#include <stdlib.h>
#include "crypto_types.h"
#include <string.h>
#include <openssl/bn.h>
#include <openssl/md5.h>

/* gem functions
* Use the modulus of x
* TODO: We may want to change this
*/
void *int_to_bn_func(int);

struct gem gem_inverse_func(struct gem *x)
{
  BIGNUM * result = BN_new();
  BN_CTX *ctx = BN_CTX_new();
  BN_mod_inverse(result, x->value, x->mod, ctx);
  BN_CTX_free(ctx);
  struct gem res;
  res.value = result;
  res.mod = x->mod;
  return res;
}

struct gem gem_add_func(struct gem* x, struct gem* y)
{
  BIGNUM *result = BN_new();
  BN_CTX *ctx = BN_CTX_new();

  BN_mod_add(result, x->value, y->value, x->mod,ctx);
  BN_CTX_free(ctx);
  struct gem res;
  res.value = result;
  res.mod = x->mod;
  return res;
}

struct gem gem_sub_func(struct gem* x, struct gem* y)
{
  BIGNUM *result = BN_new();
  BN_CTX *ctx = BN_CTX_new();

  BN_mod_sub(result, x->value, y->value, x->mod,ctx);
  BN_CTX_free(ctx);
  struct gem res;
  res.value = result;
  res.mod = x->mod;
  return res;
}

struct gem gem_mult_func(struct gem* x, struct gem* y)
{
  BIGNUM *result = BN_new();
  BN_CTX *ctx = BN_CTX_new();

  BN_mod_mul(result, x->value, y->value, x->mod,ctx);
  BN_CTX_free(ctx);
  struct gem res;
  res.value = result;
  res.mod = x->mod;
  return res;
}

struct gem gem_div_func(struct gem *x, struct gem *y)
{
  BIGNUM *result =  BN_new();
  // BIGNUM *rem = BN_new();
  BN_CTX *ctx = BN_CTX_new();
  struct gem inv = gem_inverse_func(y);
  BN_mod_mul(result,  x->value, inv.value, x->mod, ctx);
  struct gem res;
  res.value = result;
  res.mod = x->mod;
  return res;
}

struct gem gem_pow_func(struct gem *x, struct gem *y)
{
  BIGNUM *result = BN_new();
  BN_CTX *ctx = BN_CTX_new();
  if (BN_is_negative((BIGNUM *)y->value)) {
      BN_mod_inverse(x->value, x->value, x->mod, ctx);
  }
  BN_mod_exp(result, x->value, y->value, x->mod, ctx);
  BN_CTX_free(ctx);
  struct gem res;
  res.value = result;
  res.mod = x->mod;
  return res;
}

//Gem coparators
int gem_eq_func(struct gem *x, struct gem *y)
{
  return (BN_cmp(x->value, y->value) == 0 ? 1 : 0);
}

int gem_neq_func(struct gem *x, struct gem *y)
{
  return (BN_cmp(x->value, y->value) == 0 ? 0 : 1);
}

int gem_le_func(struct gem *x, struct gem *y)
{
  // fprintf(stdout, "x val: ");
  // BN_print_fp(stdout, x->value);
  // fprintf(stdout, "\ny val: ");
  // BN_print_fp(stdout, y->value);
  // fprintf(stdout, "\n");
  return (BN_cmp(x->value, y->value) < 0 ? 1 : 0);
}

int gem_ge_func(struct gem *x, struct gem *y)
{
  return (BN_cmp(x->value, y->value) >0 ? 1 : 0);
}

int gem_leq_func(struct gem *x, struct gem *y)
{
  return !gem_ge_func(x,y);
}

int gem_geq_func(struct gem *x, struct gem *y)
{
  return !gem_le_func(x, y);
}



//Int, Gem math
struct gem int_gem_add_func(void *x, struct gem *y)
{
  BIGNUM *result = BN_new();
  BN_CTX *ctx = BN_CTX_new();
  BIGNUM *bn_val = x;
  BN_mod_add(result, bn_val, y->value, y->mod, ctx);
  BN_CTX_free(ctx);
  struct gem res;
  res.value = result;
  res.mod = y->mod;
  return res;
}

struct gem int_gem_sub_func(void *x, struct gem *y)
{
  BIGNUM *result = BN_new();
  BN_CTX *ctx = BN_CTX_new();
  BIGNUM *bn_val = x;
  BN_mod_sub(result, bn_val, y->value, y->mod, ctx);
  BN_CTX_free(ctx);
  struct gem res;
  res.value = result;
  res.mod = y->mod;
  return res;
}

struct gem int_gem_mult_func(void *x, struct gem *y)
{
  BIGNUM *result = BN_new();
  BN_CTX *ctx = BN_CTX_new();
  BIGNUM *bn_val = x;
  BN_mod_mul(result, bn_val, y->value, y->mod, ctx);
  BN_CTX_free(ctx);
  struct gem res;
  res.value = result;
  res.mod = y->mod;
  return res;
}

struct gem int_gem_div_func(void *x, struct gem *y)
{
  BIGNUM *result = BN_new();
  BIGNUM *rem = BN_new();
  BN_CTX *ctx = BN_CTX_new();
  BIGNUM *bn_val = x;
  BN_div(result, rem, bn_val, y->value, ctx);
  BN_CTX_free(ctx);
  struct gem res;
  res.value = result;
  res.mod = y->mod;
  return res;
}

struct gem int_gem_pow_func(void *x, struct gem *y)
{
  BIGNUM *result = BN_new();
  BN_CTX *ctx = BN_CTX_new();
  BIGNUM *bn_val = x;
  if (BN_is_negative((BIGNUM *)y->value)) {
      BN_mod_inverse(y->value, y->value, y->mod, ctx);
  }
  BN_mod_exp(result, y->value, bn_val, y->mod, ctx);
  BN_CTX_free(ctx);
  struct gem res;
  res.value = result;
  res.mod = y->mod;
  return res;
}

// Int, Gem comparators
int int_gem_eq_func(void *x, struct gem *y)
{
  BIGNUM *bn_val = x;
  return (BN_cmp(bn_val, y->value) == 0 ? 1 : 0);
}

int int_gem_neq_func(void *x, struct gem *y)
{
  BIGNUM *bn_val = x;
  return (BN_cmp(bn_val, y->value) == 0 ? 0 :1);
}

int int_gem_less_func(void *x, struct gem *y)
{
  BIGNUM *bn_val = x;
  return (BN_cmp(bn_val, y->value) <0 ? 1 :0);
}

int int_gem_greater_func(void *x, struct gem *y)
{
  BIGNUM *bn_val = x;
  return (BN_cmp(bn_val, y->value) > 0 ? 1 :0);
}

int int_gem_leq_func(void *x, struct gem *y)
{
  return !int_gem_greater_func(x, y);
}

int int_gem_geq_func(void *x, struct gem *y)
{
  return !int_gem_less_func(x, y);
}

//Functions to create a new gem
void *int_to_bn_func(int mod)
{
  char buf[25];
  sprintf(buf, "%d", mod);
  BIGNUM *result = BN_new();
  BN_dec2bn(&result, buf);
  // printf("int_to bn: %d\n", mod);
  // BN_print_fp(stdout, result );
  // printf("\n");
  return result;
}

//Int, Int -> Gem
void * gem_mod_create_func(int val, int mod)
{
  BIGNUM *result = BN_new();
  BIGNUM *bn_val = int_to_bn_func(val);
  // printf("bn_val: %d\n", val);
  // BN_print_fp(stdout, bn_val);
  // printf("\n");
  BIGNUM *bn_mod = int_to_bn_func(mod);
  BN_CTX *ctx = BN_CTX_new();
  BN_mod(result, bn_val, bn_mod, ctx);
  if (BN_is_negative(result)) {
    BN_add(result, result, bn_mod);
  }
  // printf("bn_mod: %d\n ", mod);
  // BN_print_fp(stdout, result);
  // printf("\n");
  BN_CTX_free(ctx);
  return result;
}

//Lat, Int -> Gem
void * gem_mod_create_latint_func(void *lat, int mod)
{
  BIGNUM *result = BN_new();
  BIGNUM *bn_val = lat;
  // printf("bn_val: %d\n", val);
  // BN_print_fp(stdout, bn_val);
  // printf("\n");
  BIGNUM *bn_mod = int_to_bn_func(mod);
  BN_CTX *ctx = BN_CTX_new();
  BN_mod(result, bn_val, bn_mod, ctx);
  if (BN_is_negative(result)) {
    BN_add(result, result, bn_mod);
  }
  BN_CTX_free(ctx);
  return result;
}

//Int, Lat -> Gem
void * gem_mod_create_intlat_func(int val, void *mod)
{
  BIGNUM *result = BN_new();
  BIGNUM *bn_val = int_to_bn_func(val);
  BIGNUM *bn_mod = mod;
  BN_CTX *ctx = BN_CTX_new();
  BN_mod(result, bn_val, bn_mod, ctx);
  if (BN_is_negative(result)) {
    BN_add(result, result, bn_mod);
  }
  BN_CTX_free(ctx);
  return result;
}

//Lat, Lat -> Gem
void * gem_mod_create_latlat_func(void *val, void *mod)
{
  BIGNUM *result = BN_new();
  BIGNUM *bn_val = val;
  BIGNUM *bn_mod = mod;
  BN_CTX *ctx = BN_CTX_new();
  BN_mod(result, bn_val, bn_mod, ctx);
  if (BN_is_negative(result)) {
    BN_add(result, result, bn_mod);
  }
  BN_CTX_free(ctx);
  return result;
}

//Gem, Int -> Gem
void * gem_int_create_func( struct gem *x, int mod)
{
  BIGNUM *result = BN_new();
  BIGNUM *bn_mod = int_to_bn_func(mod);
  BN_CTX *ctx = BN_CTX_new();
  BN_mod(result, x->value, bn_mod, ctx);
  if (BN_is_negative(result)) {
    BN_add(result, result, bn_mod);
  }
  BN_CTX_free(ctx);
  return result;
}

//Int, Gem -> Gem
void *int_gem_create_func(int val, struct gem *x)
{
  BIGNUM *result = BN_new();
  BIGNUM *bn_val = int_to_bn_func(val);
  BN_CTX *ctx = BN_CTX_new();
  BN_mod(result, bn_val, x->mod, ctx);
  if (BN_is_negative(result)) {
    BN_add(result, result, x->mod);
  }
  BN_CTX_free(ctx);
  return result;
}

//Get gem
void *gem_gem(struct gem g, int index)
{
  return index ? g.value : g.mod;
}


//Printing function
void print_gem_func(struct gem x) {
  printf("%s\n", BN_bn2dec(x.value));

}

// =================== LATTICE OPS ===================

// Gem -> Lattice
void *gem_to_lat_func(struct gem *x)
{
  return x->value;
}

//String -> Lattice
void *lat_str_create_func(char *x)
{
  BIGNUM *result = BN_new();
  BN_dec2bn(&result, x);
  return result;
}

// (Lat, Lat) ops

void *lat_add_func(void *x, void *y)
{
  BIGNUM *result = BN_new();
  BN_add(result, x, y);
  return result;
}

void *lat_sub_func(void *x, void *y)
{
  BIGNUM *result = BN_new();
  BN_sub(result, x, y);
  return result;
}

void *lat_mult_func(void *x, void *y)
{
  BIGNUM *result = BN_new();
  BN_CTX *ctx = BN_CTX_new();

  BN_mul(result, x, y,ctx);
  BN_CTX_free(ctx);
  return result;
}

void *lat_div_func(void *x, void *y)
{
  BIGNUM *result =  BN_new();
  BIGNUM *rem = BN_new();
  BN_CTX *ctx = BN_CTX_new();

  BN_div(result, rem, x, y, ctx);
  return result;
}

void *lat_pow_func(void *x, void *y)
{
  BIGNUM *result = BN_new();
  BN_CTX *ctx = BN_CTX_new();
  BN_exp(result, x, y, ctx);
  BN_CTX_free(ctx);
  return result;
}

void *lat_mod_func(void *x, void *y)
{
  BIGNUM *result =  BN_new();
  BN_CTX *ctx = BN_CTX_new();

  BN_mod(result, x, y, ctx);
  return result;
}

//Lat coparators
int lat_eq_func(void *x, void *y)
{
  return (BN_cmp(x, y) == 0 ? 1 : 0);
}

int lat_neq_func(void *x, void *y)
{
  return (BN_cmp(x, y) == 0 ? 0 : 1);
}

int lat_le_func(void *x, void *y)
{
  return (BN_cmp(x, y) < 0 ? 1 : 0);
}

int lat_ge_func(void *x, void *y)
{
  return (BN_cmp(x, y) > 0 ? 1 : 0);
}

int lat_leq_func(void *x, void *y)
{
  return !lat_ge_func(x,y);
}

int lat_geq_func(void *x, void *y)
{
  return !lat_le_func(x, y);
}

// Printing a Lat
void print_lat_func(void *lat) {
  printf("%s\n", BN_bn2dec(lat));
}

/* Hash MD5 function */

char *hash_md5_func(char *str) {
  unsigned char digest[16];
  MD5_CTX mtx;

  MD5_Init(&mtx);
  MD5_Update(&mtx, str, strlen(str));
  MD5_Final(digest, &mtx);

  char mdString[33];
  for (int i = 0; i < 16; i++)
      sprintf(&mdString[i*2], "%02x", (unsigned int)digest[i]);

  BIGNUM * hash = BN_new();
  BN_CTX *ctx = BN_CTX_new();
  BN_hex2bn(&hash, mdString);

  char *ret = BN_bn2dec(hash);

  return ret;
}
