(* Ocamllex scanner for Cryptal *)

{ open Parser
  module B = Buffer}
(* Deleted BITXOR, BITOR, BITAND *)
rule token = parse
  [' ' '\t' '\r' '\n'] { token lexbuf } (* Whitespace *)
| "/*"     { comment lexbuf }           (* Comments *)
| '('      { LPAREN }
| ')'      { RPAREN }
| '{'      { LBRACE }
| '}'      { RBRACE }
| ';'      { SEMI }
| ','      { COMMA }
| '+'      { PLUS }
| '-'      { MINUS }
| '*'      { TIMES }
| "**"     { POWER	}
| '/'      { DIVIDE }
| '%'      { MOD }
| '='      { ASSIGN }
| "=="     { EQ }
| "!="     { NEQ }
| '<'      { LT }
| "<="     { LEQ }
| ">"      { GT }
| ">="     { GEQ }
| "&&"     { AND }
| "||"     { OR }
| "!"      { NOT }
| "if"     { IF }
| "else"   { ELSE }
| "for"    { FOR }
| "while"  { WHILE }
| "return" { RETURN }
| "int"    { INT }
| "gem"    { GEM }
| "uint"   { UINT }
| "lat"    { LATTICE }
| "char"   { CHAR }
| "NULL"   { NULL }
| "bool"   { BOOL }
| "void"   { VOID }
| "true"   { TRUE }
| "false"  { FALSE }
| "'"      { CHARLIT (read_char lexbuf) }
| '"'      { STRING (build_str (B.create 100) lexbuf) }
| ['0'-'9']+ as lxm { LITERAL(int_of_string lxm) }
| ['a'-'z' 'A'-'Z']['a'-'z' 'A'-'Z' '0'-'9' '_']* as lxm { ID(lxm) }
| eof { EOF }
| _ as char { raise (Failure("illegal character " ^ Char.escaped char)) }

and comment = parse
  "*/" { token lexbuf }
| _    { comment lexbuf }

and build_str sb = parse
 | '"'      { B.contents sb }
 | '\\''\\' { B.add_char sb '\\';  build_str sb lexbuf }
 | '\\''"'  { B.add_char sb '"';   build_str sb lexbuf }
 | '\\''\'' { B.add_char sb '\'';  build_str sb lexbuf }
 | '\\''n'  { B.add_char sb '\n';  build_str sb lexbuf }
 | '\\''r'  { B.add_char sb '\r';  build_str sb lexbuf }
 | '\\''t'  { B.add_char sb '\t';  build_str sb lexbuf }
 | _ as t   { B.add_char sb t;     build_str sb lexbuf }

 and read_char = parse
 | '\\''\\' { check_length '\'' lexbuf }
 | '\\''"'  { check_length '"' lexbuf }
 | '\\''\'' { check_length '\'' lexbuf }
 | '\\''n'  { check_length '\n' lexbuf }
 | '\\''r'  { check_length '\r' lexbuf }
 | '\\''t'  { check_length '\t' lexbuf }
 | '\\''0'  { check_length (char_of_int 0) lexbuf } (* '\0' char in C *)
 (* this should only match characters of length 1, as expected *)
 | (_ as t) { check_length t lexbuf }


and check_length buf = parse
 | '\'' { buf }
 | _ as t { raise( Failure ("illegal char literal " ^ Char.escaped t)) }
